import { Observable, Observer, Subject, interval } from 'rxjs';
import { share, distinctUntilChanged, takeWhile } from 'rxjs/operators';
import { WebSocketSubject, WebSocketSubjectConfig } from 'rxjs/webSocket';
import { RPCMessage } from '../../../core/type';

export interface RxWebsocketSubjectConfig<T = RPCMessage> extends WebSocketSubjectConfig<T> {
  reconnectInterval?: 5000;
  reconnectRetry?: 10;
  compressionThreshold?: 1024;
}

const DEFAULT_RX_WEBSOCKET_CONFIG: RxWebsocketSubjectConfig<any> = {
  url: '',
  deserializer: (e: MessageEvent) => e,
  serializer: (value: any) => value,
  openObserver: {
    next: (e: Event) => {
      this._connectionObserver.next(true);
    }
  },
  closeObserver: {
    next: (e: CloseEvent) => {
      this._socket = null;
      this._connectionObserver.next(false);
    }
  },
  reconnectInterval: 5000,
  reconnectRetry: 10,
  compressionThreshold: 1024,
};

export class RxWebsocketSubject extends Subject<RPCMessage> {
  private _reconnectionObservable: Observable<number>;
  private _wsSubjectConfig: RxWebsocketSubjectConfig;
  private _socket: WebSocketSubject<RPCMessage>;
  private _connectionObserver: Observer<boolean>;
  private _connectionStatus: Observable<boolean>;
  private _queryString: string;

  public constructor(private _config: RxWebsocketSubjectConfig) {
    super();

    this._connectionStatus = new Observable<boolean>((observer) => {
      this._connectionObserver = observer;
    }).pipe(
      share(),
      distinctUntilChanged(),
    );

    this._wsSubjectConfig = this._config = { ...DEFAULT_RX_WEBSOCKET_CONFIG };
    for (const key in _config) {
      if (_config.hasOwnProperty(key)) {
        switch (key) {
          case 'openObserver':
            const oo = _config[key];
            this._wsSubjectConfig[key] = {
              next: (e: Event) => {
                this._connectionObserver.next(true);
                oo.next(e);
              }
            };

            break;
          case 'closeObserver':
            const co = _config[key];
            this._wsSubjectConfig[key] = {
              next: (e: CloseEvent) => {
                this._socket = null;
                this._connectionObserver.next(false);
                co.next(e);
              }
            };

            break;
          default:
            this._wsSubjectConfig[key] = _config[key];
            break;
        }
      }
    }

    this._connectionStatus.subscribe((isConnected: boolean) => {
      if (!this._reconnectionObservable && typeof(isConnected) === 'boolean' && !isConnected) {
        this.reconnect();
      }
    });
  }

  public set queryString(query: string) {
    this._queryString = query;
  }

  public get queryString(): string | undefined {
    return this._queryString;
  }

  public get connectionStatus(): Observable<boolean> {
    return this._connectionStatus;
  }

  public connect(): void {
    const wsSubjectConfig = Object.assign({}, this._wsSubjectConfig);
    if (undefined !== this._queryString) {
      wsSubjectConfig.url = wsSubjectConfig.url + '?' + this._queryString;
    }

    this._socket = new WebSocketSubject<RPCMessage>(wsSubjectConfig);
    this._socket.subscribe(
      (m: RPCMessage) => {
        this.next(m);
      },
      (e: Event) => {
        if (!this._socket) {
          this.reconnect();
        }
      }
    );
  }

  public disconnect(): void {
    this._socket.complete();
  }

  private reconnect(): void {
    this._reconnectionObservable = interval(this._config.reconnectInterval).pipe(
      takeWhile((v, index) => {
        return index < this._config.reconnectRetry && !this._socket;
    }));

    this._reconnectionObservable.subscribe(
      () => {
        this.connect();
      },
      null,
      () => {
        this._reconnectionObservable = null;
        if (!this._socket) {
          this.complete();
          this._connectionObserver.complete();
        }
      }
    );
  }

  public write(data: RPCMessage): void {
    this._socket.next(data);
  }
}
